#!/usr/local/bin/python2.6

#
# Copyright (C) 2009, Oy L M Ericsson Ab, NomadicLab <pubsub@nomadiclab.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2 as
# published by the Free Software Foundation.
#
# Alternatively, this software may be distributed under the terms of the BSD
# license.
#
# See LICENSE and COPYING for more details.
#

"""PSIRP Local Area Intradomain Rendezvous Daemon."""


from psirp.libpsirp_py_ import *

from select import * # kevents
import errno, re, time, traceback
import array, socket # XXX
#import pdb # debugger, add breakpoint: pdb.set_trace()


ROOT_SCOPE_SID        = buffer(PSIRP_ID_LEN*'\x00')
DEFAULT_IPC_SID       = buffer(psirp_py_atoid("::bb00"))
DEFAULT_IPC_LOCAL_RID = buffer(psirp_py_atoid("::bb01"))
DEFAULT_IPC_NET_RID   = buffer(psirp_py_atoid("::bb02"))
EMPTY_FID             = buffer(PSIRP_ID_LEN*'\x00')
#FULL_FID              = buffer(PSIRP_ID_LEN*'\xff')

DEFAULT_LOCAL_SID_PATTERN = "(?!^::$)" # Everything else except "::" (Scope 0)

SOCKET_IPC = True


class EventHandler(object):
    def handle_event(self, event, pub):
        raise NotImplementedError()


class Subscriber(object):
    
    class Subs(EventHandler):
        
        def __init__(self, parent):
            self.parent = parent # Subscriber
            self.net_subs = SIdRIdDict()
            self.non_subs = SIdRIdDict(count=True)
        
        def init(self, events):
            psirp_py_debug_enter()
            
            FILENAME = "/pubsub/subs"
            subs_file = open(FILENAME)
            
            # Pseudo-publication for getting events from the subs file
            class SubsPub(object): #(Publication):
                
                def __init__(self, subs_file):
                    #super(type(self), self).__init__(None)
                    
                    self.subs_file = subs_file
                    self.pep = ps_event_page()
                    
                    self.free   = self.__del__
                    self.buffer = buffer("")
                    self.data   = None
                    self.len    = 0
                    self.rid    = buffer(PSIRP_ID_LEN*"\xff")
                    self.type   = None
                    self.addr   = 0
                    
                def __del__(self):
                    try:
                        self.subs_file.close()
                    except:
                        pass
                    
                @property
                def fd(self):
                    return self.subs_file.fileno()
                
                def read(self):
                    return psirp_py_read_event_page(self.fd, self.pep)
            
            events.register(SubsPub(subs_file),
                            self.handle_event,
                            True, True, KQ_NOTE_WRITE | KQ_NOTE_ATTRIB) # XXX
        
        def handle_event(self, event, pub):
            psirp_py_debug_enter()
            
            subs = pub.read() # This "publication" actually wraps a file
            psirp_py_debug_print(PSIRP_DBG_GARB, "%d subscriptions" % len(subs))
            
            for sid, rid in subs:
                # Make copies, beacause data in the underlying page can change
                sid, rid = copybuf(sid), copybuf(rid)
                sid_str = psirp_idtoa(sid)
                
                psirp_py_debug_print(PSIRP_DBG_GARB,
                                     "Subscription to SId/RId: %s/%s" \
                                         % (sid_str, psirp_idtoa(rid)))
                
                if self.non_subs.contains(sid, rid):
                    self.non_subs.remove(sid, rid)
                    psirp_py_debug_print(PSIRP_DBG_GARB,
                                         "Subscription ignored "
                                         "(self-inflicted)")
                    continue # for sid, rid in subs
                
                elif self.net_subs.contains(sid, rid):
                    psirp_py_debug_print(PSIRP_DBG_GARB,
                                         "Subscription ignored "
                                         "(already subscribed)")
                    # Should we support re-subscribing?
                    continue # for sid, rid in subs

                elif re.match(self.parent.sid_pattern, sid_str) is None:
                    psirp_py_debug_print(PSIRP_DBG_GARB,
                                         "Subscription ignored "
                                         "(%s is not a network scope)" \
                                             % sid_str)
                    continue # for sid, rid in subs

                else:
                    try:
                        self.non_subs.add(sid, rid)
                        pub = psirp_py_subscribe(sid, rid)
                        
                        psirp_py_debug_print(PSIRP_DBG_GARB,
                                             "Subscription ignored "
                                             "(already stored locally)")
                        continue # for sid, rid in subs
                    except EnvironmentError, e:
                        if psirp_py_notfound(e):
                            # Publication not available locally
                            pass
                        else:
                            raise
                    
                    if self.parent.net.metadata.contains(sid, rid):
                        self.net_subs.add(sid, rid)
                        psirp_py_debug_print(PSIRP_DBG_INFO,
                                             "Already have metadata, "
                                             "subscribing to data, RFD=%s" \
                                                 % psirp_idtoa(rid))
                        self.parent.net.publish_ipc_md_ext(
                            PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                            self.parent.net.metadata[sid][rid],
                            EMPTY_FID,
                            False)
                    else:
                        self.net_subs.add(sid, rid)
                        psirp_py_debug_print(
                            PSIRP_DBG_INFO,
                            "Subscription stored, SId/RId=%s" \
                                % psirp_idstoa(sid, rid))
                        
                        # First we need to subscribe to the metadata
                        psirp_py_debug_print(PSIRP_DBG_INFO,
                                             "Subscribing to metadata "
                                             "from the RZV node"
                                             ", RId=%s" \
                                                 % psirp_idtoa(rid))
                        self.parent.net.publish_ipc_md(
                            PSIRP_HDR_RZV_SUBSCRIBE_METADATA,
                            Metadata(sid, rid))
        
    class Net(EventHandler):

        def __init__(self, parent, ipc_sid, ipc_local_rid, ipc_net_rid):
            self.parent = parent # Subscriber
            self.ipc_sid = ipc_sid
            self.ipc_local_rid = ipc_local_rid
            self.ipc_net_rid = ipc_net_rid
            
            self.metadata = SIdRIdDict()
    
        def init(self, events):
            psirp_py_debug_enter()
            
            if (SOCKET_IPC):
                ipc_net_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                ipc_net_sock.bind(("localhost", 0xacdc))
                
                # Pseudo-publication for getting events from an IPC socket
                class SockPub(object): #(Publication):
                    
                    def __init__(self, sock):
                        #super(type(self), self).__init__(None)
                        
                        self.sock = sock
                        
                        self.free   = self.__del__
                        self.buffer = buffer("")
                        self.data   = None
                        self.len    = 0
                        self.rid    = buffer(PSIRP_ID_LEN*"\xff")
                        self.type   = None
                        self.addr   = 0
                        
                    def __del__(self):
                        try:
                            self.sock.close()
                        except:
                            pass
                    
                    @property
                    def fd(self):
                        return self.sock.fileno()
                    
                    def read(self, len=128):
                        return buffer(self.sock.recv(len))
                
                events.register(SockPub(ipc_net_sock),
                                self.handle_event,
                                False, False, 0,
                                KQ_FILTER_READ)
                return
            # else: ...
            
            # Subscribe to network-triggered events (inbound in this context)
            try:
                self.parent.subs.non_subs.add(self.ipc_sid,
                                              self.ipc_net_rid)
                ipc_pub = psirp_py_subscribe(self.ipc_sid, self.ipc_net_rid)
            except EnvironmentError, e:
                if psirp_py_notfound(e):
                    # The IPC publication didn't exist yet, so we can
                    # try to publish it.
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "IPC publication does not yet exist, "
                                         "creating it")
                    
                    createpub(self.ipc_sid, self.ipc_net_rid)
                    time.sleep(0.5) # Wait for scoped to finish
                    
                    self.parent.subs.non_subs.add(self.ipc_sid,
                                                  self.ipc_net_rid)
                    ipc_pub = psirp_py_subscribe(self.ipc_sid, self.ipc_net_rid)
                else:
                    raise
            
            events.register(ipc_pub, self.handle_event)
        
        def handle_event(self, event, ipc_pub_old):
            psirp_py_debug_enter()
            
            if (SOCKET_IPC):
                # XXX
                metadata = psirp_py_buf_hdrs_metadata(ipc_pub_old.read())
            else:
                self.parent.subs.non_subs.add(self.ipc_sid, self.ipc_net_rid)
                ipc_pub_new = psirp_py_subscribe(self.ipc_sid, self.ipc_net_rid)
                self.parent.events.replace(ipc_pub_old,
                                           ipc_pub_new, self.handle_event)
                ipc_pub_old = None
                metadata = psirp_py_pub_hdrs_metadata(ipc_pub_new.pub)
                
            sid  = copybuf(metadata.getsid())
            rid  = copybuf(metadata.getrid())
            # Collected FId to the pub'r or sub'r:
            fid  = copybuf(metadata.getfid())
            len  = metadata.len
            type = metadata.rzvhdr_type
            
            rid_str = psirp_idtoa(rid)
            
            psirp_py_debug_print(PSIRP_DBG_INFO,
                                 "Received metadata:\n"
                                 "         SId/RId: %s\n"
                                 "         FId:     %s\n"
                                 "         len:     %d\n"
                                 "         type:    0x%02x" \
                                     % (psirp_idstoa(sid, rid),
                                        psirp_idtoa(fid),
                                        len,
                                        type))
            
            if type == PSIRP_HDR_RZV_PUBLISH_METADATA:
                # Somebody in the network published something,
                # or we have subscribed to a publication.
                psirp_py_debug_print(PSIRP_DBG_GARB,
                                     "Received publish metadata")
                
                # We could also subscribe:
                try:
                    self.parent.subs.non_subs.add(sid, rid)
                    pub = psirp_py_subscribe(sid, rid)
                    
                    # We already have this publication. Perhaps somebody
                    # else published a new version or something, but in
                    # this implementation we currently just drop the metadata.
                    return
                except EnvironmentError, e:
                    if psirp_py_notfound(e):
                        # OK, we can proceed since we don't seem to have it yet.
                        pass
                    else:
                        raise
                
                # Add to metadata map
                md = Metadata(sid, rid, fid, len)
                self.metadata.add(sid, rid, md)
                psirp_py_debug_print(PSIRP_DBG_INFO,
                                     "Metadata stored, RId=%s" % rid_str);
                
                # If we have this in our subscription map, we have
                # previously subscribed to it and want to get the data.
                if self.parent.subs.net_subs.contains(sid, rid):
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "Subscribe to data, RId=%s" \
                                             % rid_str);
                    
                    self.publish_ipc_md_ext(
                        PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                        md,
                        EMPTY_FID,
                        False)
            
            elif type == PSIRP_HDR_RZV_SUBSCRIBE_METADATA:
                # Somebody in the network subscribed to metadata
                psirp_py_debug_print(PSIRP_DBG_GARB,
                                     "Recevied subscribe metadata")
                
                if self.metadata.contains(sid, rid):
                    # We have the corresponding metadata, so let's
                    # publish it
                    md = self.metadata[sid][rid]
                    
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "Publishing metadata "
                                         "to the subscriber, "
                                         "RId=%s" % rid_str);
                    
                    md_out = Metadata(sid, rid, fid, md.len)
                    self.publish_ipc_md(PSIRP_HDR_RZV_PUBLISH_METADATA,
                                        md_out)
                else:
                    # Check if we have the publication locally
                    try:
                        self.parent.subs.non_subs.add(sid, rid)
                        spub = psirp_py_subscribe(sid, rid)
                    except EnvironmentError, e:
                        if psirp_py_notfound(e):
                            psirp_py_debug_print(
                                PSIRP_DBG_INFO,
                                "Subscription to unknown publication")
                            # XXX: In order to support sub-before-pub
                            # cases we should also store subscriptions
                            # that came from the net, especially if we
                            # are a local RZV node...
                            return
                        else:
                            raise
                    
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "Publishing metadata "
                                         "to the subscriber, "
                                         "RId=%s" % rid_str);
                    
                    md_out = Metadata(sid, rid, fid, spub.len)
                    self.publish_ipc_md(PSIRP_HDR_RZV_PUBLISH_METADATA,
                                        md_out)
            
            elif type == PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK:
                # Somebody in the network subscribed to data
                psirp_py_debug_print(PSIRP_DBG_GARB,
                                     "Received subscribe data")
                
                if self.metadata.contains(sid, rid):
                    # We have the corresponding metadata, but
                    # presumably not the data. We pass on this
                    # subscription to the publisher.
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "Relay subscription to data chunks")
                    md = self.metadata[sid][rid]
                    self.publish_ipc_md_ext(PSIRP_HDR_RZV_SUBSCRIBE_DATACHUNK,
                                            md,
                                            fid,
                                            True)
                else:
                    # Check if we have this publication locally. If we have it,
                    # send it out.
                    try:
                        self.parent.subs.non_subs.add(sid, rid)
                        pub = psirp_py_subscribe(sid, rid)
                        
                        md = Metadata(sid, rid, fid, pub.len)
                        psirp_py_debug_print(PSIRP_DBG_INFO,
                                             "Publish data chunks:\n"
                                             "         SId/RId: %s\n"
                                             "         FId:     %s\n"
                                             "         len:     %d"
                                             % (psirp_idstoa(sid, rid),
                                                psirp_idtoa(fid),
                                                pub.len))
                        
                        self.publish_ipc_md(PSIRP_HDR_RZV_PUBLISH_DATACHUNK, md)
                    except EnvironmentError, e:
                        if psirp_py_notfound(e):
                            psirp_py_debug_print(
                                PSIRP_DBG_INFO,
                                "Subscription to unknown publication")
                        else:
                            raise
            
            elif type == PSIRP_HDR_RZV_PUBLISH_DATACHUNK:
                psirp_py_debug_print(PSIRP_DBG_WARN,
                                     "Received publish datachunk "
                                     "-- unexpectedly and unrequestedly!")
            
            else:
                psirp_py_debug_print(PSIRP_DBG_WARN,
                                     "Unknown IPC publication type: %d" % type)
                return
        
        def publish_ipc_md(self, type, md):
            psirp_py_debug_enter()
            
            if (SOCKET_IPC): # XXX
                buf = array.array('c', 128*'\x00')
                metadata = psirp_py_buf_hdrs_metadata(buffer(buf))
                pub = None
            else:
                pub = Publication(psirp_py_create_md_pub(None))
                metadata = psirp_py_pub_hdrs_metadata(pub.pub)
                buf = pub.buffer
        
            metadata.rzvhdr_type = type
            metadata.hdr_type = PSIRP_HDR_MD
            metadata.setsid(md.sid)
            metadata.setrid(md.rid)
            metadata.setfid(md.fid)
            metadata.len = md.len
            
            if (SOCKET_IPC): # XXX
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.sendto(buffer(buf), ("localhost", 0xdcac))
                return
            #else: ...
                
            psirp_py_publish(self.ipc_sid, self.ipc_local_rid, pub)
            psirp_py_debug_print(PSIRP_DBG_FNCE, "RETURN")
        
        def publish_ipc_md_ext(self, type, md, extra_fid, relay):
            psirp_py_debug_enter()
            
            if (SOCKET_IPC): # XXX
                buf = array.array('c', 161*'\x00')
                metadata = psirp_py_buf_hdrs_metadata_ext(buffer(buf))
                pub = None
            else:
                pub = Publication(psirp_py_create_md_ext_pub(None))
                metadata = psirp_py_pub_hdrs_metadata_ext(pub.pub)
                buf = pub.buffer
                
            metadata.md_hdr.rzvhdr_type = type
            metadata.md_hdr.hdr_type = PSIRP_HDR_MD
            metadata.setmdsid(md.sid)
            metadata.setmdrid(md.rid)
            metadata.setmdfid(md.fid)
            metadata.md_hdr.len = md.len
            
            metadata.setextfid(extra_fid)
            metadata.md_ext.relay = relay
            
            if (SOCKET_IPC): # XXX
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.sendto(buffer(buf), ("localhost", 0xdcac))
                return
            #else: ...
            
            psirp_py_publish(self.ipc_sid, self.ipc_local_rid, pub)
            psirp_py_debug_print(PSIRP_DBG_FNCE, "RETURN")
            
    
    class Scope(EventHandler):
        
        def __init__(self, parent, parent_sid, sid):
            self.parent = parent
            self.parent_sid = parent_sid
            self.sid = sid
        
        def init(self, events, net=False):
            psirp_py_debug_enter()
            
            # Subscribe to scope
            psirp_py_debug_print(PSIRP_DBG_GARB, "Subscribe to scope")
            
            self.parent.subs.non_subs.add(self.sid, self.sid)
            scope_pub = psirp_py_subscribe(self.sid, self.sid)
            
            psirp_py_debug_print(PSIRP_DBG_GARB,
                                 "Read RIds in scope with SId: %s" \
                                     % psirp_idtoa(self.parent_sid));
            scope_data = psirp_py_scope(scope_pub.pub)
            
            # Read and process initial scope contents
            for rid in scope_data:
                self.init_pub(events, rid, net)
        
        def init_pub(self, events, rid, net=False):
            rid_str = psirp_idtoa(rid)
            psirp_py_debug_print(PSIRP_DBG_GARB,
                                 "Subscribe to RId: %s" % rid_str)
            
            # Subscribe to this publication
            self.parent.subs.non_subs.add(self.sid, rid)
            pub = psirp_py_subscribe(self.sid, rid)
            
            # Register to its events
            if pub.type == PSIRP_PUB_SCOPE:
                if rid == ROOT_SCOPE_SID \
                        or re.match(self.parent.sid_pattern,
                                    rid_str) is not None:
                    if rid == self.sid:
                        psirp_py_debug_print(PSIRP_DBG_INFO,
                                             "Registering to scope events, "
                                             "SId=%s" % rid_str)
                        events.register(pub, self.handle_event)
                    else:
                        psirp_py_debug_print(PSIRP_DBG_GARB,
                                             "Initializing subscope, "
                                             "SId=%s" % rid_str)
                        subscope = self.parent.Scope(self.parent,
                                                     self.sid, rid)
                        subscope.init(events, net=True)
                else:
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "SId=%s is not a network scope" \
                                             % rid_str)
            else:
                psirp_py_debug_print(PSIRP_DBG_INFO,
                                     "Registering to publication events, "
                                     "RId=%s" % rid_str)
                events.register(pub, self.handle_event_pub)
                
                if net:
                    if self.parent.net.metadata.contains(self.sid, rid):
                        # Remove metadata
                        self.parent.net.metadata.remove(self.sid, rid)
                        
                    if self.parent.subs.net_subs.contains(self.sid, rid):
                        # This publication had been subscribed previously.
                        # Remove the pending subscription.
                        self.parent.subs.net_subs.remove(self.sid, rid)
                    else:
                        # Issue IPC command to get new metadata published
                        # into the network
                        psirp_py_debug_print(PSIRP_DBG_INFO,
                                             "Publishing metadata "
                                             "to the RZV node, "
                                             "RId=%s" % rid_str)
                        md = Metadata(self.sid, rid, EMPTY_FID, pub.len)
                        self.parent.net.publish_ipc_md(
                            PSIRP_HDR_RZV_PUBLISH_METADATA, md)
        
        def handle_event(self, event, scope_pub_old):
            psirp_py_debug_enter()
            
            if event.fflags & NOTE_PUBLISH:
                # Something was published in the scope
                self.parent.subs.non_subs.add(self.sid, self.sid)
                scope_pub_new = psirp_py_subscribe(self.sid, self.sid)
                
                scope_data_old = set(psirp_py_scope(scope_pub_old.pub))
                scope_data_new = set(psirp_py_scope(scope_pub_new.pub))
                
                for rid in scope_data_new.difference(scope_data_old):
                    psirp_py_debug_print(PSIRP_DBG_INFO,
                                         "New scope entry: SId/RId=%s" \
                                             % psirp_idstoa(self.sid, rid))
                    self.init_pub(self.parent.events, rid, net=True)
                
                # XXX: in the future, items might also have been removed
                
                self.parent.events.replace(scope_pub_old,
                                           scope_pub_new, self.handle_event)
                scope_pub_old = None
            else:
                psirp_py_debug_print(PSIRP_DBG_WARN, "Scope subscribed?")
        
        def handle_event_pub(self, event, pub_old):
            psirp_py_debug_enter()
            
            if event.fflags & NOTE_PUBLISH:
                rid = copybuf(pub_old.rid)
                
                psirp_py_debug_print(PSIRP_DBG_INFO,
                                     "Publication update, RId=%s" \
                                         % psirp_idtoa(rid))
                
                # Issue IPC command to get new metadata published
                # into the network
                self.parent.subs.non_subs.add(self.sid, rid)
                pub_new = psirp_py_subscribe(self.sid, rid)
                md = Metadata(self.parent_sid, rid, EMPTY_FID, pub_new.len)
                self.parent.net.publish_ipc_md(
                    PSIRP_HDR_RZV_PUBLISH_METADATA, md)
                self.parent.events.replace(pub_old,
                                           pub_new, self.handle_event_pub)
        
    
    def __init__(self, sid_pattern, ipc_sid, ipc_local_rid, ipc_net_rid):
        self.sid_pattern = re.compile(sid_pattern)
        
        self.events = Events()
        self.subs = self.Subs(self)
        self.net = self.Net(self, ipc_sid, ipc_local_rid, ipc_net_rid)
        self.root_scope = self.Scope(self, ROOT_SCOPE_SID, ROOT_SCOPE_SID)
    
    def laird_listen(self):
        psirp_py_debug_enter()
        
        self.subs.init(self.events)
        self.net.init(self.events)
        self.root_scope.init(self.events)
        
        # Listen
        psirp_py_debug_print(PSIRP_DBG_INFO, "Listening to events")
        try:
            self.events.listen()
        except:
            psirp_py_debug_print(PSIRP_DBG_ERROR,
                                 "Rendezvous listener:\n%s" \
                                     % traceback.format_exc())
        
        # Cleanup
        psirp_py_debug_print(PSIRP_DBG_GARB, "Cleaning up")
        self.events.cleanup()
        
        psirp_py_debug_print(PSIRP_DBG_GARB, "Done")

    
def createpub(sid, rid):
        pub = psirp_py_create(1)
        psirp_py_publish(sid, rid, pub)
        return pub
    

def copybuf(buf):
        return buffer(buf[:])
    

class SIdRIdDict(dict):
    # {SId<buffer>: {RId<buffer>: counter<int>|item<object>, ...}, ...}
    
    def __init__(self, count=False):
        self.count = count
    
    def add(self, sid, rid, item=None):
        if sid not in self:
            self[sid] = {}
        if self.count and rid in self[sid]:
            self[sid][rid] += 1
        else:
            if item is None:
                item = 1
            self[sid][rid] = item
    
    def remove(self, sid, rid):
        if sid in self and rid in self[sid]:
            if self.count and self[sid][rid] > 1:
                self[sid][rid] -= 1
            else:
                del self[sid][rid]
            if len(self[sid]) == 0:
                del self[sid]
    
    def contains(self, sid, rid):
        return (self[sid][rid] if sid in self and rid in self[sid] else 0)


class Metadata(object):
    def __init__(self, sid, rid, fid=EMPTY_FID, len=0):
        self.sid = sid
        self.rid = rid
        self.fid = fid
        self.len = len


class Events(dict):
    # {ident: (pub, evh), ...}
        
    def __init__(self, kq=None):
        self.kq = kqueue() if kq is None else kq
    
    def register(self, pub, evh, reg_p=True, reg_s=False, fflags=0,
                 filter=KQ_FILTER_VNODE):
        # XXX: pub -> fd+udata?        
        psirp_py_debug_enter()
        
        ident, udata = pub.fd, pub.addr
        
        fflags |= \
            (NOTE_PUBLISH if reg_p else 0) \
            | (NOTE_SUBSCRIBE if reg_s else 0)
        #fflags |= NOTE_UNMAP # XXX: ?
        
        evf = kevent(ident,
                     filter,
                     KQ_EV_ADD | KQ_EV_CLEAR,
                     fflags,
                     0,
                     udata)
        self.kq.control([evf], 0, None)
        
        self[ident] = [pub, evh]
    
    def unregister(self, pub):
        ident = pub.fd
        
        evf = kevent(ident,
                     (KQ_FILTER_READ if pub.__class__.__name__ == "SockPub" #XXX
                      else KQ_FILTER_VNODE),
                     KQ_EV_DELETE)
        try:
            self.kq.control([evf], 0, None)
        except:
            psirp_py_debug_print(PSIRP_DBG_WARN,
                                 "unregister failed (%d)" % ident) # XXX
        
        if ident in self:
            del self[ident]
    
    def replace(self, pub_old, pub_new, evh_new):
        # XXX: What if there are pending events for pub_old?
        self.unregister(pub_old)
        self.register(pub_new, evh_new)
    
    def cleanup(self):
        for pub, evh in self.values():
            self.unregister(pub)
        self.kq.close()
        
        self.clear()
    
    def print_pub_map(self):
        pub_map_str = "Publication map (%d):\n" % len(self)
        for ident in self.keys():
            pub, evh = self[ident]
            pub_map_str += "         %d: %s\n" % (ident, psirp_idtoa(pub.rid))
        psirp_py_debug_print(PSIRP_DBG_GARB, "%s" % pub_map_str)
    
    def listen(self):
        psirp_py_debug_enter()
        evl = None
        try:
            while True:
                psirp_py_debug_print(PSIRP_DBG_GARB, "Listening to events...")
                evl = self.kq.control(None, 1, None) # listen to kevents
                psirp_py_debug_print(PSIRP_DBG_GARB, "%s" % evl)
                
                if evl is not None:
                    for ev in evl:
                        if ev.ident in self:
                            pub, evh = self[ev.ident]
                            try:
                                evh(ev, pub) # call event handler
                            except Exception, e: # XXX
                                psirp_py_debug_print(PSIRP_DBG_ERR,
                                    "Event handler:\n%s" \
                                        % traceback.format_exc())
                        else:
                            psirp_py_debug_print(PSIRP_DBG_WARN,
                                                 "Unknown ident: %d" % ev.ident)
                
                evl = None
        except (KeyboardInterrupt, SystemExit):
            psirp_py_debug_print(PSIRP_DBG_INFO,
                                 "Interrupted%s" \
                                     % ("" if evl is None else ": %s" % evl))



def main():
    import getopt
        
    try:
        opts, args = getopt.gnu_getopt(sys.argv[1:], "s:r:q:t:ch", [])
        #print(opts, args)
    except getopt.GetoptError, err:
        print(str(err))
        sys.exit(1)
    
    debug_level = 0
    debug_colors = False
    sid_pattern = DEFAULT_LOCAL_SID_PATTERN
    ipc_sid = DEFAULT_IPC_SID
    ipc_local_rid = DEFAULT_IPC_LOCAL_RID
    ipc_net_rid = DEFAULT_IPC_NET_RID
    
    for o, a in opts:
        if o == "-h":
            print("%s %s" \
                      % (str(__file__).split("/")[-1],
                         "[-s sid_pattern -r ipc_local_rid -q ipc_net_rid "
                         "-t debug_level -c -h]"))
            sys.exit()
        elif o == "-r":
            ipc_local_rid = psirp_py_atoid(a)
        elif o == "-q":
            ipc_net_rid = psirp_py_atoid(a)
        elif o == "-t":
            debug_level = int(a)
        elif o == "-c":
            debug_colors = True
        else:
            assert False, "Unhandled option: %r" % o

    psirp_py_debug_init(debug_level, debug_colors)
    
    kwargs = {"sid_pattern": sid_pattern,
              "ipc_sid": ipc_sid,
              "ipc_local_rid": ipc_local_rid,
              "ipc_net_rid": ipc_net_rid}
    
    s = Subscriber(**kwargs)
    s.laird_listen()


if __name__ == "__main__":
    main()
